/*
 * feedbackd
 *
 * Copyright (c) 2012 - 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <dlfcn.h>
#include <assert.h>
#include <vconf.h>
#include <time.h>
#include <string.h>
#include <libsyscommon/dbus-system.h>

#include "core/log.h"
#include "core/list.h"
#include "core/common.h"
#include "core/device-idler.h"
#include "core/config-parser.h"
#include "haptic.h"

#define FEEDBACK_BASE_PATH         "/usr/share/feedback/"
#define STANDARD_FILE_PATH         FEEDBACK_BASE_PATH"vibration/"
#define VIBRATION_CONF_PATH        FEEDBACK_BASE_PATH"vibration.conf"

#define HAPTIC_CONF_PATH                "/etc/feedbackd/haptic.conf"
#define SIGNAL_CHANGE_HARDKEY           "ChangeHardkey"
#define SIGNAL_POWEROFF_STATE           "ChangeState"
#define SIGNAL_VIBRATOR_INITIATED       "InitiateVibrator"

/* hardkey vibration variable */
#define HARDKEY_VIB_ITERATION		1
#define HARDKEY_VIB_FEEDBACK		3
#define HARDKEY_VIB_DURATION		30
#define HAPTIC_FEEDBACK_STEP		20
#define DEFAULT_FEEDBACK_LEVEL		3
#define HIGH_FEEDBACK_LEVEL		100
#define LOW_FEEDBACK_LEVEL		0

/* power on, power off vibration variable */
#define POWER_ON_VIB_DURATION			300
#define POWER_OFF_VIB_DURATION			300
#define POWER_VIB_FEEDBACK			100

#define MAX_EFFECT_BUFFER			(64*1024)

#ifndef VCONFKEY_RECORDER_STATE
#define VCONFKEY_RECORDER_STATE "memory/recorder/state"
#define VCONFKEY_RECORDER_STATE_RECORDING	2
#endif

#define CHECK_VALID_OPS(ops, r)		((ops) ? true : !(r = -ENODEV))
#define RETRY_CNT	3

#define INTENSITY_BASE_RATE        (10000)
#define VALUE_MAX_LEN              10
#define VIB_LOCK_TIMEOUT_MAX       (300000) /* 5minutes */

/*
		1,A_W or A,A_W or 250D250W250D750W

		85,10000,
		90,0,
		105,10000,
		0,end
*/
struct vibration_config {
	char *pattern; /* pattern name */
	char *standard; /* assigned standard pattern name */
	dd_list *data; /* duration_data list */
	unsigned int pattern_duration;
	int unlimit;
};

struct duration_data {
	int duration;
	int intensity;
	int wait;
};

enum poweroff_type {
	POWER_OFF_NONE = 0,
	POWER_OFF_POPUP,
	POWER_OFF_DIRECT,
	POWER_OFF_RESTART,
};

struct haptic_info {
	char *sender;
	dd_list *handle_list;
	guint id_watch;
};

struct vibrate_pattern_info {
	unsigned int handle;
	char *pattern;
	int level;
	int priority;
};

struct vibrate_monotone_info {
	unsigned int handle;
	int duration;
	int level;
	int priority;
};

/* for playing */
static int g_handle;

/* pattern configuration list */
static dd_list *vib_conf_list;

guint duration_timer;

/* haptic operation variable */
static dd_list *h_head;
static dd_list *haptic_handle_list;
static const struct haptic_plugin_ops *h_ops;
static enum haptic_type h_type;
static bool haptic_disabled;

struct haptic_config {
	int level;
	int *level_arr;
	int sound_capture;
};

static struct haptic_config haptic_conf;

static int haptic_start(void);
static int haptic_stop(void);
static int haptic_internal_init(void);
static int remove_haptic_info(struct haptic_info *info);

struct haptic_data cur_h_data;

void add_haptic(const struct haptic_ops *ops)
{
	DD_LIST_APPEND(h_head, (void *)ops);
}

void remove_haptic(const struct haptic_ops *ops)
{
	DD_LIST_REMOVE(h_head, (void *)ops);
}

static int insert_conf_data(dd_list **conf_data, struct duration_data *update)
{
	struct duration_data *data;

	data = (struct duration_data *)calloc(1, sizeof(struct duration_data));
	if (!data) {
		_E("Not enough memory.");
		return -ENOMEM;
	}
	memcpy(data, update, sizeof(struct duration_data));
	/* insert vibration pattern data */
	// to debug :
	// _D("%dD%dI%dF%dO%dW", data->duration, data->intensity, data->frequency, data->overdriving, data->wait);
	DD_LIST_APPEND(*conf_data, data);
	return 0;
}

static void get_pattern_property(char **iter, char property, int *value)
{
	char *check;
	unsigned long int val;

	check = strchr(*iter, property);
	if (!check)
		return;

	*check = '\0';
	val = strtoul(*iter, NULL, 10);
	if (errno == EINVAL || errno == ERANGE) {
		val = 0;
		_E("Failed to get value of %s: %d", *iter, errno);
	}
	if (val > VIB_LOCK_TIMEOUT_MAX)
		val = VIB_LOCK_TIMEOUT_MAX;

	*value = (int)val;

	*iter = check + 1;
}

/* [A]xxxDxxxIxxxFxxxOxxxW format */
static int insert_raw_data_format(dd_list **conf_data, char *value)
{
	struct duration_data update = {0, };
	char *iter;
	char *end;
	int pattern_duration = 0;

	if (!value)
		return insert_conf_data(conf_data, &update);
	if (!conf_data) {
		_E("Invalid parameter: Configuration list is null");
		return -EINVAL;
	}

	iter = value;
	end = iter + strlen(iter);
	while (iter < end) {
		memset(&update, 0, sizeof(struct duration_data));

		get_pattern_property(&iter, 'D', &update.duration);
		get_pattern_property(&iter, 'I', &update.intensity);
		if (update.intensity > INTENSITY_BASE_RATE)
			update.intensity = INTENSITY_BASE_RATE;
		get_pattern_property(&iter, 'W', &update.wait);

		if (update.duration == 0 && update.wait == 0) {
			_D("Pattern duration is zero.");
			break;
		}

		pattern_duration += (update.duration + update.wait);
		if (pattern_duration > VIB_LOCK_TIMEOUT_MAX) {
			_D("Max pattern duration.");
			pattern_duration = VIB_LOCK_TIMEOUT_MAX;
		}

		if (insert_conf_data(conf_data, &update) < 0)
			return -EINVAL;
	}

	return pattern_duration;
}

static int get_config_data(int count, bool *packed, char *val, unsigned int *pattern_duration, struct duration_data *update)
{
	static int duration = 0;
	int value;

	if (count > 2 || count <= 0) {
		_E("Invalid parameter: count is invalid");
		return -EINVAL;
	}

	if (!packed || !val || !pattern_duration || !update) {
		_E("Invalid parameter: %p %p %p %p", packed, val, pattern_duration, update);
		return -EINVAL;
	}

	get_pattern_property(&val, 'C', &value);

	if (count == 1) {
		duration = value;
		*pattern_duration += duration;
		if (*pattern_duration > VIB_LOCK_TIMEOUT_MAX) {
			_D("Max pattern duration.");
			*pattern_duration = VIB_LOCK_TIMEOUT_MAX;

		}
		return count;
	}

	if (value > INTENSITY_BASE_RATE)
		value = INTENSITY_BASE_RATE;
	if (*packed == false) {
		update->duration = duration;
		update->intensity = value;
		*packed = true;
		return 0;
	}
	if (value == 0)
		update->wait = duration;
	else
		update->wait = 0;
	*packed = false;
	duration = 0;

	return -1;
}

/*
	duration, intensity, frequency, overdriving
	waiting duration, intensity=0, frequency, overdriving
	85,10000,
	90,0,
	105,10000,
	0,end
*/
static int load_standard_format(const char *pattern)
{
	struct duration_data update = {0, };
	bool packed = false;
	struct vibration_config *conf;
	int ret = 0, count = 0, end = 2;
	int index = 0;
	int fd;
	char path[PATH_MAX], elem, val[VALUE_MAX_LEN] = {0, };

	snprintf(path, sizeof(path), STANDARD_FILE_PATH"%s", pattern);
	fd = open(path, O_RDONLY);
	if (fd < 0)
		return -ENOENT;

	conf = (struct vibration_config *)calloc(1, sizeof(struct vibration_config));
	if (!conf) {
		_E("Failed to alloc.");
		ret = -errno;
		goto error_out;
	}

	conf->pattern = strdup(pattern);
	if (!conf->pattern) {
		_E("Failed to copy pattern data(%s).", pattern);
		ret = -errno;
		goto error_out;
	}

	/* make feedback pattern(xDxIxFxOxW) format from d,i,f,o, */
	while (read(fd, &elem, 1) != 0) {
		if (end == 0) {
			if (insert_conf_data(&conf->data, &update) < 0)
				goto error_out;
			break;
		}
		if (elem == 'e' || elem == 'n' || elem == 'd') {
			end--;
			continue;
		}
		if (elem == '\n') {
			count = 0;
			continue;
		}
		if (elem == ',') {
			count++;
			val[index] = 'C';
			index = 0;

			ret = get_config_data(count, &packed, val, &(conf->pattern_duration), &update);
			if (ret < 0) {
				if (ret == -EINVAL)
					break;
				if (insert_conf_data(&conf->data, &update) < 0)
					goto error_out;
				memset(&update, 0, sizeof(struct duration_data));
			} else
				count = ret;
		} else {
			if (index < (VALUE_MAX_LEN - 2)) /* Temporal limit */
				val[index++] = elem;
			else
				_E("Pattern(%s) is out of bound: %s", pattern, val);
		}
	}
	close(fd);
	DD_LIST_APPEND(vib_conf_list, conf);
	return ret;

error_out:
	if (fd >= 0)
		close(fd);
	if (conf) {
		if (conf->pattern)
			free(conf->pattern);
		free(conf);
	}
	return -ENOENT;
}

static int vibration_load_config(struct parse_result *result, void *user_data)
{
	struct vibration_config *conf;
	char *value;
	char *check;
	int len;
	int duration;

	if (!result)
		return 0;

	if (!MATCH(result->section, "Vibration"))
		return 0;


	if (!result->name || !result->value)
		return 0;

	conf = (struct vibration_config *)calloc(1, sizeof(struct vibration_config));
	if (!conf) {
		_E("Failed to alloc.");
		return -ENOMEM;
	}

	conf->pattern_duration = 0;
	conf->pattern = strdup(result->name);
	if (!conf->pattern) {
		_E("Failed to copy pattern data(%s).", result->name);
		goto error_out;
	}

	value = result->value;
	len = strlen(value);

	if (len == 0) {
		if (insert_raw_data_format(&conf->data, NULL) < 0)
			goto error_out;
		DD_LIST_APPEND(vib_conf_list, conf);
		return 0;
	}

	/* Load Standard Pattern Name */
	/* value: 1,A_W or A,A_W */
	if ((check = strchr(value, ','))) {
		*check = '\0';
		if (strncmp(value, "A", 1) == 0)
			conf->unlimit = 1;
		value = check + 1;
		conf->standard = strdup(value);
		if (!conf->standard) {
			_E("Failed to copy standard name.");
			goto error_out;
		}
		DD_LIST_APPEND(vib_conf_list, conf);
		return 0;
	}

	/* value : A100D or 100D0W or 250D250W250D750W */
	/* Load Vibration Pattern Data */
	check = strchr(value, 'A');
	if (check) {
		*check = '\0';
		conf->unlimit = 1;
		if (!value)
			len = len - 1;
		else
			len = len - strlen(value) - 1;
		value = check + 1;
	}
	duration = insert_raw_data_format(&conf->data, value);
	if (duration < 0) {
		conf->pattern_duration = 0;
		goto error_out;
	} else
		conf->pattern_duration = duration;
	DD_LIST_APPEND(vib_conf_list, conf);

	return 0;

error_out:
	if (conf) {
		if (conf->pattern)
			free(conf->pattern);
		if (conf->standard)
			free(conf->standard);
	}
	return -ENOMEM;
}

static void load_standard_vibration_patterns(void)
{
	DIR *dir;
	struct dirent *dent;
	int ret;

	dir = opendir(STANDARD_FILE_PATH);
	if (!dir) {
		_E("Failed to load '%s' Use default value.", STANDARD_FILE_PATH);
		return;
	}
	while ((dent = readdir(dir))) {
		if (dent->d_type == DT_DIR)
			continue;
		ret = load_standard_format(dent->d_name);
		if (ret < 0)
			_E("Failed to parse %s: %d", dent->d_name, ret);
	}
	closedir(dir);
	_D("Success to load '%s'", STANDARD_FILE_PATH);
}

void pattern_config_parse(void)
{
	int ret;

	ret = config_parse(VIBRATION_CONF_PATH, vibration_load_config, NULL);
	if (ret < 0)
		_E("Failed to load '%s'. Use default value: %d", VIBRATION_CONF_PATH, ret);

	load_standard_vibration_patterns();
}

static int haptic_module_load(void)
{
	struct haptic_ops *ops;
	dd_list *elem;
	int r;

	pattern_config_parse();
	/* find valid plugin */
	DD_LIST_FOREACH(h_head, elem, ops) {
		if (ops->is_valid && ops->is_valid()) {
			if (ops->load)
				h_ops = ops->load();
			h_type = ops->type;
			break;
		}
	}

	if (!CHECK_VALID_OPS(h_ops, r)) {
		_E("Can't find the valid haptic device.");
		return r;
	}

	/* solution bug
	 * we do not use internal vibration except power off.
	 * if the last handle is closed during the playing of vibration,
	 * solution makes unlimited vibration.
	 * so we need at least one handle. */
	haptic_internal_init();

	return 0;
}

static int convert_magnitude_by_conf(int level)
{
	int i, step;

	if (level < LOW_FEEDBACK_LEVEL)
		level = LOW_FEEDBACK_LEVEL;
	else if (level > HIGH_FEEDBACK_LEVEL)
		level = HIGH_FEEDBACK_LEVEL;

	step = 100 / (haptic_conf.level-1);
	for (i = 0; i < haptic_conf.level; ++i) {
		if (level <= i*step) {
			_D("Level changed. %d -> %d", level, haptic_conf.level_arr[i]);
			return haptic_conf.level_arr[i];
		}
	}

	_D("Play default level.");
	return DEFAULT_FEEDBACK_LEVEL * HAPTIC_FEEDBACK_STEP;
}

GVariant *hdbus_get_count(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret, val;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	ret = h_ops->get_device_count(&val);
	if (ret >= 0)
		ret = val;

exit:
	return g_variant_new("(i)", ret);
}

void haptic_name_owner_changed(GDBusConnection *connection,
	const gchar     *name,
	gpointer         user_data)
{
	dd_list *n;
	struct haptic_info *info = user_data;
	int handle;

	_I("%s (sender=%s)", __func__, name);

	if (!info)
		return;

	for (n = info->handle_list; n; n = n->next) {
		handle = (int)(long)n->data;
		h_ops->stop_device(handle);
		h_ops->close_device(handle);
	}

	remove_haptic_info(info);
}

static struct haptic_info *add_haptic_info(const char *sender)
{
	struct haptic_info *info;

	assert(sender);

	info = calloc(1, sizeof(struct haptic_info));
	if (!info)
		return NULL;

	info->sender = strdup(sender);
	if (!info->sender) {
		free(info);
		return NULL;
	}

	DD_LIST_APPEND(haptic_handle_list, info);

	info->id_watch = dbus_handle_watch_name(sender, NULL, haptic_name_owner_changed, info, NULL);

	return info;
}

static int remove_haptic_info(struct haptic_info *info)
{
	assert(info);

	dbus_handle_unwatch_name(info->id_watch);

	DD_LIST_REMOVE(haptic_handle_list, info);
	DD_LIST_FREE_LIST(info->handle_list);
	if (info->sender) {
		free(info->sender);
	}
	free(info);

	return 0;
}

static struct haptic_info *get_matched_haptic_info(const char *sender)
{
	dd_list *n;
	struct haptic_info *info;
	int len;

	assert(sender);

	len = strlen(sender) + 1;
	DD_LIST_FOREACH(haptic_handle_list, n, info) {
		if (!strncmp(info->sender, sender, len))
			return info;
	}

	return NULL;
}

GVariant *hdbus_open_device(GDBusConnection *conn,
							const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
							GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)

{
	int index, handle, ret;
	struct haptic_info *info;

	/* Load haptic module before booting done */
	if (!CHECK_VALID_OPS(h_ops, ret)) {
		ret = haptic_module_load();
		if (ret < 0)
			goto exit;
	}

	g_variant_get(param, "(i)", &index);

	ret = h_ops->open_device(index, &handle);
	if (ret < 0)
		goto exit;

	if (!sender) {
		ret = -EPERM;
		h_ops->close_device(handle);
		goto exit;
	}

	info = get_matched_haptic_info(sender);
	if (!info) {
		info = add_haptic_info(sender);
		if (!info) {
			_E("Failed to create haptic information.");
			ret = -EPERM;
			h_ops->close_device(handle);
			goto exit;
		}
	}

	DD_LIST_APPEND(info->handle_list, (gpointer)(long)handle);

	ret = handle;

exit:
	return g_variant_new("(i)", ret);
}

int clear_current_data(void)
{
	cur_h_data.handle = INVALID_HANDLE;
	cur_h_data.priority = PRIORITY_MIN;

	if (duration_timer) {
		_I("Remove duration_timer.");
		g_source_remove(duration_timer);
		duration_timer = 0;
	}

	return 0;
}

GVariant *hdbus_close_device(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int handle;
	int ret;
	struct haptic_info *info;
	int cnt;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	g_variant_get(param, "(u)", &handle);

	if (!sender) {
		_E("Failed to get sender from dbus message.");
		ret = -EPERM;
		goto exit;
	}

	ret = h_ops->close_device(handle);
	if (ret < 0)
		goto exit;

	if (cur_h_data.handle == handle) {
		/* Remove duration_timer for vibrate_effect */
		clear_current_data();
	}

	info = get_matched_haptic_info(sender);
	if (!info) {
		_E("Failed to find the matched haptic info.");
		goto exit;
	}

	DD_LIST_REMOVE(info->handle_list, (gpointer)(long)handle);
	cnt = DD_LIST_LENGTH(info->handle_list);
	if (cnt == 0)
		remove_haptic_info(info);

exit:
	return g_variant_new("(i)", ret);
}

static gboolean _cb(void *data)
{
	if (duration_timer) {
		g_source_remove(duration_timer);
		duration_timer = 0;
	}

	cur_h_data.handle = INVALID_HANDLE;
	cur_h_data.priority = PRIORITY_MIN;

	return G_SOURCE_REMOVE;
}

static void vibrate_monotone_idler_cb(void *data)
{
	struct vibrate_monotone_info *vibrate_info;
	int ret;

	if (!data)
		return;

	vibrate_info = (struct vibrate_monotone_info *)data;

	if (vibrate_info->priority < cur_h_data.priority) {
		_I("Handle(%d) skip low priority. pre=%d now=%d", vibrate_info->handle, cur_h_data.priority, vibrate_info->priority);
		free(vibrate_info);
		return;
	}

	if (duration_timer) {
		g_source_remove(duration_timer);
		duration_timer = 0;
	}

	cur_h_data.vibration_data = NULL;
	cur_h_data.handle = vibrate_info->handle;
	cur_h_data.level = vibrate_info->level;
	cur_h_data.priority = vibrate_info->priority;
	cur_h_data.stop = false;
	cur_h_data.unlimit = false;

	ret = device_power_request_lock(POWER_LOCK_CPU, vibrate_info->duration);
	if (ret != DEVICE_ERROR_NONE)
		_E("Failed to request power lock.");

	duration_timer = g_timeout_add(vibrate_info->duration, _cb, NULL);
	h_ops->vibrate_monotone(cur_h_data.handle, vibrate_info->duration, cur_h_data.level, cur_h_data.priority);

	free(vibrate_info);
}

GVariant *hdbus_vibrate_monotone(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	struct vibrate_monotone_info *vibrate_info;
	unsigned int handle;
	int duration, level, priority, ret = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	if (haptic_disabled)
		goto exit;

	g_variant_get(param, "(uiii)", &handle, &duration, &level, &priority);

	/* convert as per conf value */
	level = convert_magnitude_by_conf(level);
	if (level < 0) {
		ret = -EINVAL;
		goto exit;
	}

	if (priority < PRIORITY_MIN)
		priority = PRIORITY_MIN;
	else if (priority > PRIORITY_TOP)
		priority = PRIORITY_TOP;

	if (duration <= 0) {
		_E("Skip vibrate handle(%d) requested less than 0.", handle);
		ret = -EINVAL;
		goto exit;
	}

	vibrate_info = calloc(1, sizeof(struct vibrate_monotone_info));
	if (!vibrate_info) {
		_E("Failed to allocate memory for vibrate_info.");
		ret = -errno;
		goto exit;
	}

	vibrate_info->handle = handle;
	vibrate_info->duration = duration;
	vibrate_info->level = level;
	vibrate_info->priority = priority;

	ret = add_idle_request(vibrate_monotone_idler_cb, (void *)vibrate_info);
exit:
	return g_variant_new("(i)", ret);
}

static gboolean haptic_duration_play(void *data)
{
	dd_list *head, *n, *next;
	struct duration_data *node;
	int level;
	int ret = 0;

	if (duration_timer) {
		g_source_remove(duration_timer);
		duration_timer = 0;
	}

	if (!data) {
		if (cur_h_data.unlimit) /* In case of unlimit pattern, do not stop */
			head = cur_h_data.vibration_data;
		else {
			cur_h_data.handle = INVALID_HANDLE;
			cur_h_data.priority = PRIORITY_MIN;
			goto out;
		}
	} else
		head = (dd_list *)data;

	if (cur_h_data.stop) {
		_I("Stop currunt vibration.");
		cur_h_data.stop = false;
		cur_h_data.handle = INVALID_HANDLE;
		cur_h_data.priority = PRIORITY_MIN;
		goto out;
	}

	DD_LIST_FOREACH_SAFE(head, n, next, node) {
		_D("Handle(%d) play=%dms and Wait=%dms %s type.",
			cur_h_data.handle, node->duration, node->wait,
			cur_h_data.unlimit ? "Unlimit" : "Once");
		if ((node->duration + node->wait) <= 0) {
			if (!cur_h_data.unlimit) {
				cur_h_data.handle = INVALID_HANDLE;
				cur_h_data.priority = PRIORITY_MIN;
				break;
			} else {
				next = cur_h_data.vibration_data;
				continue;
			}
		}

		if (node->intensity)
			level = (cur_h_data.level * node->intensity) / INTENSITY_BASE_RATE;
		else
			level = cur_h_data.level;

		duration_timer = g_timeout_add((node->duration + node->wait), haptic_duration_play, (void *)next);

		ret = h_ops->vibrate_monotone(cur_h_data.handle, node->duration, level, cur_h_data.priority);

		break;
	}
	if (ret != 0) {
		_D("Auto stop vibration.");
		cur_h_data.stop = true;
	}
out:
	return G_SOURCE_REMOVE;
}

static void vibrate_pattern_idler_cb(void *data)
{
	struct vibrate_pattern_info *vibrate_info;
	dd_list *elem;
	struct vibration_config *conf;
	char pattern[PATH_MAX];
	int ret;
	int unlimit = 0;

	if (!data)
		return;

	vibrate_info = (struct vibrate_pattern_info *)data;

	if (!(vibrate_info->pattern)) {
		free(vibrate_info);
		return;
	}

	/* Same or higher priority pattern should be played */
	if (vibrate_info->priority < cur_h_data.priority) {
		_I("Handle(%d) skip low priority. pre=%d now=%d", vibrate_info->handle, cur_h_data.priority, vibrate_info->priority);
		goto out;
	}

	snprintf(pattern, sizeof(pattern), "%s", vibrate_info->pattern);
	DD_LIST_FOREACH(vib_conf_list, elem, conf) {
		if (!conf->pattern)
			continue;
		if (strcmp(conf->pattern, pattern))
			continue;
		if (conf->standard) {
			unlimit = conf->unlimit;
			snprintf(pattern, sizeof(pattern), "%s", conf->standard);
			continue;
		}

		if (unlimit)
			cur_h_data.unlimit = unlimit;
		else
			cur_h_data.unlimit = conf->unlimit;
		cur_h_data.vibration_data = conf->data;
		cur_h_data.handle = vibrate_info->handle;
		cur_h_data.level = vibrate_info->level;
		cur_h_data.priority = vibrate_info->priority;
		cur_h_data.stop = false;
		_I("Handle(%d) play=%s pri=%d %s", cur_h_data.handle, conf->pattern, cur_h_data.priority,
				cur_h_data.unlimit ? "Unlimit" : "Once");

		if (conf->pattern_duration <= 0)
			break;

		if (!cur_h_data.unlimit && conf->pattern_duration > 0) {
			ret = device_power_request_lock(POWER_LOCK_CPU, (int)conf->pattern_duration);
			if (ret != DEVICE_ERROR_NONE)
				_E("Failed to request power lock.");
		}
		haptic_duration_play((void *)cur_h_data.vibration_data);
		goto out;
	}
	_E("Handle(%d) %s is not supported.", vibrate_info->handle, pattern);

out:
	free(vibrate_info->pattern);
	free(vibrate_info);
}

GVariant *hdbus_vibrate_pattern(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	struct vibrate_pattern_info *vibrate_info;
	unsigned int handle;
	char *pattern = NULL;
	int level, priority, ret = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	if (haptic_disabled)
		goto exit;

	g_variant_get(param, "(usii)",  &handle, &pattern, &level, &priority);

	/* convert as per conf value */
	level = convert_magnitude_by_conf(level);
	if (level < 0) {
		ret = -EINVAL;
		goto exit;
	}
	if (priority < PRIORITY_MIN)
		priority = PRIORITY_MIN;
	else if (priority > PRIORITY_TOP)
		priority = PRIORITY_TOP;

	vibrate_info = calloc(1, sizeof(struct vibrate_pattern_info));
	if (!vibrate_info) {
		_E("Failed to allocate memory for vibrate_info.");
		ret = -errno;
		goto exit;
	}
	vibrate_info->handle = handle;
	vibrate_info->pattern = pattern;
	pattern = NULL;
	if (!vibrate_info->pattern) {
		_E("Failed to allocate memory for pattern.");
		ret = -errno;
		free(vibrate_info);
		goto exit;
	}
	vibrate_info->level = level;
	vibrate_info->priority = priority;

	ret = add_idle_request(vibrate_pattern_idler_cb, (void *)vibrate_info);

exit:
	g_free(pattern);
	return g_variant_new("(i)", ret);
}

GVariant *hdbus_stop_device(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	unsigned int handle;
	int ret = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	if (haptic_disabled) {
		_I("Haptic disabled.");
		goto exit;
	}

	g_variant_get(param, "(u)", &handle);

	if (cur_h_data.handle != handle) {
		_D("Not the request from current vibration handle(%d). Skip.", handle);
		goto exit;
	}

	ret = h_ops->stop_device(handle);

	/* Remove duration_timer for vibrate_effect */
	clear_current_data();

exit:
	return g_variant_new("(i)", ret);
}

GVariant *hdbus_show_handle_list(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	dd_list *n;
	dd_list *elem;
	struct haptic_info *info;

	_D("Sender Handle");
	DD_LIST_FOREACH(haptic_handle_list, n, info) {
		for (elem = info->handle_list; elem; elem = elem->next)
			_D("%s %d", info->sender, (int)(long)elem->data);
	}

	return g_variant_new_tuple(NULL, 0);
}

int pattern_is_supported(const char *pattern)
{
	dd_list *elem;
	struct vibration_config *conf;
	int ret;

	if (!pattern)
		return -EINVAL;

	ret = 0;
	DD_LIST_FOREACH(vib_conf_list, elem, conf) {
		if (!conf->pattern)
			continue;
		if (!strcmp(conf->pattern, pattern)) {
			ret = true;
			break;
		}
	}

	return ret;
}

GVariant *hdbus_pattern_is_supported(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)

{
	char *data = NULL;
	int ret = 0;

	if (!CHECK_VALID_OPS(h_ops, ret))
		goto exit;

	if (haptic_disabled)
		goto exit;

	g_variant_get(param, "(s)", &data);

	ret = pattern_is_supported(data);

	_I("%s is supported: %d", data, ret);

exit:
	g_free(data);
	return g_variant_new("(i)", ret);
}

static int haptic_internal_init(void)
{
	int r;
	if (!CHECK_VALID_OPS(h_ops, r))
		return r;
	return h_ops->open_device(HAPTIC_MODULE_DEVICE_ALL, &g_handle);
}

static int haptic_internal_exit(void)
{
	int r;
	if (!CHECK_VALID_OPS(h_ops, r))
		return r;
	return h_ops->close_device(g_handle);
}

static void haptic_hardkey_changed_cb(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	int level, status, ret;

	if (!CHECK_VALID_OPS(h_ops, ret)) {
		ret = haptic_module_load();
		if (ret < 0)
			return;
	}

	if (!g_handle)
		haptic_internal_init();

	/* if haptic is stopped, do not play vibration */
	if (haptic_disabled)
		return;

	if (vconf_get_bool(VCONFKEY_SETAPPL_VIBRATION_STATUS_BOOL, &status) < 0) {
		_E("Failed to get VCONFKEY_SETAPPL_VIBRATION_STATUS_BOOL.");
		status = 1;
	}

	/* when turn off haptic feedback option */
	if (!status)
		return;

	ret = vconf_get_int(VCONFKEY_SETAPPL_TOUCH_FEEDBACK_VIBRATION_LEVEL_INT, &level);
	if (ret < 0) {
		_E("Failed to get VCONFKEY_SETAPPL_TOUCH_FEEDBACK_VIBRATION_LEVEL_INT.");
		level = HARDKEY_VIB_FEEDBACK;
	}

	ret = h_ops->vibrate_monotone(g_handle, HARDKEY_VIB_DURATION,
			level*HAPTIC_FEEDBACK_STEP, PRIORITY_HIGH);
	if (ret < 0)
		_E("Failed to vibrate buffer: %d", ret);

	return;
}

static void haptic_poweroff_cb(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	int type = POWER_OFF_NONE;
	int ret;
	struct timespec time = {0,};

	g_variant_get(param, "(i)", &type);

	if (type != POWER_OFF_DIRECT && type != POWER_OFF_RESTART)
		return;

	_D("Poweroff: %d", type);

	if (!CHECK_VALID_OPS(h_ops, ret)) {
		ret = haptic_module_load();
		if (ret < 0)
			return;
	}

	if (!g_handle)
		haptic_internal_init();

	/* power off vibration */
	_I("Handle=%d dur=%dms pri=%d level=%d",
		g_handle, POWER_OFF_VIB_DURATION, PRIORITY_HIGH, POWER_VIB_FEEDBACK);
	ret = h_ops->vibrate_monotone(g_handle, POWER_OFF_VIB_DURATION,
			POWER_VIB_FEEDBACK, PRIORITY_HIGH);
	if (ret < 0) {
		_E("Failed to vibrate_monotone: %d", ret);
		return;
	}

	/* sleep for vibration */
	time.tv_nsec = POWER_OFF_VIB_DURATION * NANO_SECOND_MULTIPLIER;
	nanosleep(&time, NULL);
	return;
}

static void sound_capturing_cb(keynode_t *key, void *data)
{
	int status;

	status = vconf_keynode_get_int(key);

	/* if sound capture is in use, this value is 1(true). */
	if (status == VCONFKEY_RECORDER_STATE_RECORDING)
		haptic_stop();
	else
		haptic_start();
}

static int parse_section(struct parse_result *result, void *user_data, int index)
{
	struct haptic_config *conf = (struct haptic_config *)user_data;

	if (!result)
		return 0;

	if (!result->section || !result->name || !result->value)
		return 0;

	if (MATCH(result->name, "sound_capture")) {
		conf->sound_capture = atoi(result->value);
	} else if (MATCH(result->name, "level")) {
		conf->level = atoi(result->value);
		if (conf->level < 0 || conf->level >= INT_MAX - 1) {
			_E("You must set level with positive number in integer range.");
			return -EINVAL;
		}
		conf->level_arr = calloc(sizeof(int), conf->level);
		if (!conf->level_arr) {
			_E("Failed to allocate memory for level.");
			return -errno;
		}
	} else if (MATCH(result->name, "value")) {
		if (index < 0)
			return -EINVAL;
		conf->level_arr[index] = atoi(result->value);
	}

	return 0;
}

static int haptic_load_config(struct parse_result *result, void *user_data)
{
	struct haptic_config *conf = (struct haptic_config *)user_data;
	char name[NAME_MAX];
	int ret;
	static int index;

	if (!result)
		return 0;

	if (!result->section || !result->name || !result->value)
		return 0;

	// Parsing 'Haptic' section
	if (MATCH(result->section, "Haptic")) {
		ret = parse_section(result, user_data, -1);
		if (ret < 0) {
			_E("Failed to parse 'Haptic' section: %d", ret);
			return ret;
		}
		goto out;
	}

	// Parsing 'level' section
	for (index = 0; index < conf->level; ++index) {
		snprintf(name, sizeof(name), "level%d", index);
		if (MATCH(result->section, name)) {
			ret = parse_section(result, user_data, index);
			if (ret < 0) {
				_E("Failed to parse 'level' section: %d", ret);
				return ret;
			}
			goto out;
		}
	}

out:
	return 0;
}

static const dbus_method_s hdbus_methods[] = {
	{ "GetCount",          NULL,   "i", hdbus_get_count },            // device_haptic_get_count
	{ "OpenDevice",         "i",   "i", hdbus_open_device },          // device_haptic_open, feedback_initialize
	{ "CloseDevice",        "u",   "i", hdbus_close_device },         // device_haptic_close, feedback_deinitialize
	{ "StopDevice",         "u",   "i", hdbus_stop_device },          // device_haptic_stop, feedback_stop
	{ "VibrateMonotone", "uiii",   "i", hdbus_vibrate_monotone },     // device_haptic_vibrate
	{ "VibratePattern",  "usii",   "i", hdbus_vibrate_pattern },      // feedback_play*
	{ "ShowHandleList",    NULL,  NULL, hdbus_show_handle_list },
	{ "IsSupported",        "s",   "i", hdbus_pattern_is_supported }, // feedback_is_supported_pattern
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.name = VIBRATOR_INTERFACE_HAPTIC,
	.methods = hdbus_methods,
	.nr_methods = ARRAY_SIZE(hdbus_methods),
};

int haptic_probe(void)
{
	/**
	 * load haptic module.
	 * if there is no haptic module,
	 * feedbackd does not activate a haptic interface.
	 */
	return haptic_module_load();
}

guint id_sig_change_hardkey;
guint id_sig_pwr_off_state;

void haptic_init(void)
{
	int r;

	/* get haptic data from configuration file */
	r = config_parse(HAPTIC_CONF_PATH, haptic_load_config, &haptic_conf);
	if (r < 0) {
		_E("Failed to load configuration file(%s): %d", HAPTIC_CONF_PATH, r);
		safe_free(haptic_conf.level_arr);
	}

	/* init dbus interface */
	r = dbus_handle_register_dbus_object(NULL, VIBRATOR_PATH_HAPTIC, &dbus_interface);
	if (r < 0)
		_E("Failed to init hdbus interface and method: %d", r);

	/* register notifier for below each event */
	id_sig_change_hardkey = subscribe_dbus_signal(NULL, DEVICED_PATH_KEY, DEVICED_INTERFACE_KEY, SIGNAL_CHANGE_HARDKEY, haptic_hardkey_changed_cb, NULL, NULL);
	if (id_sig_change_hardkey <= 0) {
		_E("Failed to register signal handler: %d", r);
		return;
	}

	id_sig_pwr_off_state = subscribe_dbus_signal(NULL, DEVICED_PATH_POWEROFF, DEVICED_INTERFACE_POWEROFF, SIGNAL_POWEROFF_STATE, haptic_poweroff_cb, NULL, NULL);
	if (id_sig_pwr_off_state <= 0) {
		_E("Failed to register signal handler: %d", r);
		return;
	}

	dbus_handle_emit_dbus_signal(NULL,
			VIBRATOR_PATH_HAPTIC,
			VIBRATOR_INTERFACE_HAPTIC,
			SIGNAL_VIBRATOR_INITIATED,
			NULL);

	/* add watch for sound capturing value */
	if (haptic_conf.sound_capture) {
		r = vconf_notify_key_changed(VCONFKEY_RECORDER_STATE, sound_capturing_cb, NULL);
		if (r != 0)
			_W("Add watch for VCONFKEY_RECORDER_STATE failed.");
	}

	/* Initialize vibration_handle (Use vibration now) */
	cur_h_data.handle = INVALID_HANDLE;
	cur_h_data.priority = PRIORITY_MIN;
}

void haptic_exit(void)
{
	struct haptic_ops *ops;
	dd_list *elem;
	int r;

	/* remove watch */
	if (haptic_conf.sound_capture) {
		r = vconf_ignore_key_changed(VCONFKEY_RECORDER_STATE, sound_capturing_cb);
		if (r != 0)
			_W("Remove watch for VCONFKEY_RECORDER_STATE failed.");
	}

	/* unregister notifier for below each event */
	unsubscribe_dbus_signal(NULL, id_sig_change_hardkey);
	unsubscribe_dbus_signal(NULL, id_sig_pwr_off_state);

	/* release haptic data memory */
	safe_free(haptic_conf.level_arr);

	if (!CHECK_VALID_OPS(h_ops, r))
		return;

	/* haptic exit for feedbackd */
	haptic_internal_exit();

	/* release plugin */
	DD_LIST_FOREACH(h_head, elem, ops) {
		if (ops->is_valid && ops->is_valid()) {
			if (ops->release)
				ops->release();
			h_ops = NULL;
			break;
		}
	}
}

static int haptic_start(void)
{
	_I("Start");
	haptic_disabled = false;
	return 0;
}

static int haptic_stop(void)
{
	_I("Stop");
	haptic_disabled = true;
	return 0;
}
