/*
 * feedbackd
 *
 * Copyright (c) 2012 - 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <linux/input.h>

#include "core/log.h"
#include "core/list.h"
#include "haptic.h"

#define MAX_MAGNITUDE			0xFFFF
#define PERIODIC_MAX_MAGNITUDE	0x7FFF	/* 0.5 * MAX_MAGNITUDE */
#define RUMBLE_MAX_MAGNITUDE	0xFFFF

#define DEV_INPUT   "/dev/input"
#define EVENT		"event"

#define MAX_DATA 16
#define FF_INFO_MAGIC 0xDEADFEED

struct ff_info_header {
	unsigned int magic;
	int iteration;
	int ff_info_data_count;
};

struct ff_info_data {
	int type;/* play, stop etc */
	int magnitude; /* strength */
	int length; /* in ms for stop, play*/
};

struct ff_info_buffer {
	struct ff_info_header header;
	struct ff_info_data data[MAX_DATA];
};

struct ff_info {
	int handle;
	guint timer;
	struct ff_effect effect;
	struct ff_info_buffer *ffinfobuffer;
	int currentindex;
};

static int ff_fd;
static dd_list *ff_list;
static dd_list *handle_list;
static char ff_path[PATH_MAX];
static int unique_number;

static int stop_device(int device_handle);

struct ff_info *read_from_list(int handle)
{
	struct ff_info *temp;
	dd_list *elem;

	DD_LIST_FOREACH(ff_list, elem, temp) {
		if (temp->handle == handle)
			return temp;
	}
	return NULL;
}

static bool check_valid_handle(struct ff_info *info)
{
	struct ff_info *temp;
	dd_list *elem;

	DD_LIST_FOREACH(ff_list, elem, temp) {
		if (temp == info)
			break;
	}

	if (!temp)
		return false;
	return true;
}

static bool check_fd(int *fd)
{
	int ffd;

	if (*fd > 0)
		return true;

	ffd = open(ff_path, O_RDWR);
	if (ffd < 0)
		return false;

	*fd = ffd;
	return true;
}

static int ff_stop(int fd, struct ff_effect *effect);
static gboolean timer_cb(void *data)
{
	struct ff_info *info = (struct ff_info *)data;

	if (!info) {
		_E("Failed to check info.");
		return G_SOURCE_REMOVE;
	}

	if (!check_valid_handle(info)) {
		_E("Failed to check valied info.");
		return G_SOURCE_REMOVE;
	}

	_I("Stop vibration by timer. id(%d)", info->effect.id);

	/* stop previous vibration */
	ff_stop(ff_fd, &info->effect);

	/* reset timer */
	info->timer = 0;

	return G_SOURCE_REMOVE;
}

static int ff_find_device(void)
{
	DIR *dir;
	struct dirent *dent;
	char ev_path[PATH_MAX];
	unsigned long features[1+FF_MAX/sizeof(unsigned long)];
	int fd, ret;

	dir = opendir(DEV_INPUT);
	if (!dir)
		return -errno;

	while (1) {
		dent = readdir(dir);
		if (dent == NULL)
			break;

		if (dent->d_type == DT_DIR ||
			!strstr(dent->d_name, "event"))
			continue;

		snprintf(ev_path, sizeof(ev_path), "%s/%s", DEV_INPUT, dent->d_name);

		fd = open(ev_path, O_RDWR);
		if (fd < 0) {
			_E("Failed to open '%s'.: %d", ev_path, errno);
			continue;
		}

		/* get force feedback device */
		memset(features, 0, sizeof(features));
		ret = ioctl(fd, EVIOCGBIT(EV_FF, sizeof(features)), features);
		if (ret == -1) {
			close(fd);
			continue;
		}

		if (test_bit(FF_CONSTANT, features))
			_D("'%s' type: constant", ev_path);
		if (test_bit(FF_PERIODIC, features))
			_D("'%s' type: periodic", ev_path);
		if (test_bit(FF_SPRING, features))
			_D("'%s' type: spring", ev_path);
		if (test_bit(FF_FRICTION, features))
			_D("'%s' type: friction", ev_path);
		if (test_bit(FF_RUMBLE, features))
			_D("'%s' type: rumble", ev_path);

		if (test_bit(FF_RUMBLE, features)) {
			memcpy(ff_path, ev_path, strlen(ev_path));
			close(fd);
			closedir(dir);
			return 0;
		}

		close(fd);
	}

	closedir(dir);
	return -1;
}

static int ff_init_effect(struct ff_effect *effect)
{
	if (!effect)
		return -EINVAL;

	/*Only rumble supported as of now*/
	effect->type = FF_RUMBLE;
	effect->replay.length = 0;
	effect->replay.delay = 10;
	effect->id = -1;
	effect->u.rumble.strong_magnitude = 0;

	return 0;
}

static int ff_set_effect(struct ff_effect *effect, int length, int level)
{
	double magnitude;

	if (!effect) {
		_E("There is no valid effect.");
		return -EINVAL;
	}

	magnitude = (double)level/HAPTIC_MODULE_FEEDBACK_MAX;
	magnitude *= RUMBLE_MAX_MAGNITUDE;

	_I("magnitude=%d length=%d", (int)magnitude, length);

	/* set member variables in effect struct */
	effect->u.rumble.strong_magnitude = (int)magnitude;
	effect->replay.length = length;		/* length millisecond */

	return 0;
}

static int ff_play(int fd, struct ff_effect *effect)
{
	struct input_event play;
	int ret;

	if (fd < 0 || !effect) {
		_E("Fd(%d) or effect(%s) is invalid", fd, effect ? "not null" : "null");
		return -EINVAL;
	}

	/* upload an effect */
	if (ioctl(fd, EVIOCSFF, effect) == -1) {
		_E("Failed to ioctl");
		return -errno;
	}

	/* play vibration*/
	play.type = EV_FF;
	play.code = effect->id;
	play.value = 1; /* 1 : PLAY, 0 : STOP */

	ret = write(fd, (const void *)&play, sizeof(play));
	if (ret == -1) {
		_E("Failed to write");
		return -errno;
	}

	return 0;
}

static int ff_stop(int fd, struct ff_effect *effect)
{
	struct input_event stop;
	int ret;

	if (fd < 0 || !effect)
		return -EINVAL;

	/* Stop vibration */
	stop.type = EV_FF;
	stop.code = effect->id;
	stop.value = 0; /* 1 : PLAY, 0 : STOP */
	ret = write(fd, (const void *)&stop, sizeof(stop));
	if (ret == -1)
		return -errno;

	/* removing an effect from the device */
	if (ioctl(fd, EVIOCRMFF, effect->id) == -1)
		return -errno;

	/* reset effect id */
	effect->id = -1;

	return 0;
}

/* START: Haptic Module APIs */
static int get_device_count(int *count)
{
	/* suppose there is just one haptic device */
	if (count)
		*count = 1;

	return 0;
}

static int open_device(int device_index, int *device_handle)
{
	struct ff_info *info;
	int n;
	bool found = false;
	dd_list *elem;

	if (!device_handle)
		return -EINVAL;

	/* if it is the first element */
	n = DD_LIST_LENGTH(ff_list);
	if (n == 0 && !ff_fd) {
		_I("First element: open ff driver");
		/* open ff driver */
		ff_fd = open(ff_path, O_RDWR);
		if (ff_fd < 0) {
			_E("Failed to open %s : %d", ff_path, errno);
			return -errno;
		}
	}

	/* allocate memory */
	info = calloc(sizeof(struct ff_info), 1);
	if (!info) {
		_E("Failed to allocate memory : %d", errno);
		return -errno;
	}

	/* initialize ff_effect structure */
	ff_init_effect(&info->effect);

	if (unique_number == INT_MAX)
		unique_number = 0;

	while (found != true) {
		++unique_number;
		elem = DD_LIST_FIND(handle_list, (gpointer)(long)unique_number);
		if (!elem)
			found = true;
	}

	info->handle = unique_number;

	/* add info to local list */
	DD_LIST_APPEND(ff_list, info);
	DD_LIST_APPEND(handle_list, (gpointer)(long)info->handle);

	*device_handle = info->handle;
	return 0;
}

static int close_device(int device_handle)
{
	struct ff_info *info;
	int n;

	info = read_from_list(device_handle);
	if (!info) {
		_E("Handle %d failed to check info", device_handle);
		return -ENOENT; /* 2 */
	}

	if (!check_valid_handle(info)) {
		_E("Handle %d failed to check valid handle", device_handle);
		return -EINVAL; /* 22 */
	}

	if (!check_fd(&ff_fd)) {
		_E("Handle %d failed to check fd", device_handle);
		return -ENODEV; /* 19 */
	}

	/* stop vibration */
	stop_device(device_handle);

	DD_LIST_REMOVE(handle_list, (gpointer)(long)info->handle);

	safe_free(info->ffinfobuffer);
	/* remove info from local list */
	DD_LIST_REMOVE(ff_list, info);
	safe_free(info);

	/* if it is the last element */
	n = DD_LIST_LENGTH(ff_list);
	if (n == 0 && ff_fd) {
		_I("Last element: close ff driver");
		/* close ff driver */
		close(ff_fd);
		ff_fd = 0;
	}

	return 0;
}

static int vibrate_monotone(int device_handle, int duration, int feedback, int priority)
{
	struct ff_info *info;
	int ret;

	info = read_from_list(device_handle);
	if (!info) {
		_E("Handle %d failed to check list", device_handle);
		return -EINVAL;
	}

	if (!check_valid_handle(info)) {
		_E("Handle %d failed to check handle", device_handle);
		return -EINVAL;
	}

	if (!check_fd(&ff_fd))
		return -ENODEV;

	if (duration <= 0) {
		_I("Handle %d skip requests with duration 0", device_handle);
		return 0;
	}

	/* Zero(0) is the infinitely vibration value */
	if (duration == HAPTIC_MODULE_DURATION_UNLIMITED)
		duration = 0;

	/* unregister existing timer */
	if (info->timer) {
		ff_stop(ff_fd, &info->effect);
		g_source_remove(info->timer);
		info->timer = 0;
	}

	/* set effect as per arguments */
	ff_init_effect(&info->effect);
	ret = ff_set_effect(&info->effect, duration, feedback);
	if (ret < 0) {
		_E("Handle %d fail to set effect(duration:%d, feedback:%d) : %d",
				device_handle, duration, feedback, ret);
		return ret;
	}

	/* play effect as per arguments */
	ret = ff_play(ff_fd, &info->effect);
	if (ret < 0) {
		_E("Handle %d fail to play haptic effect(fd:%d id:%d) : %d",
				device_handle, ff_fd, info->effect.id, ret);
		return ret;
	}

	/* register timer */
	if (duration) {
		info->timer = g_timeout_add(duration, timer_cb, info);
		if (!info->timer)
			_E("Handle %d failed to add timer callback", device_handle);
	}

	_D("Handle %d effect id : %d %dms", device_handle, info->effect.id, duration);

	return 0;
}

static int stop_device(int device_handle)
{
	struct ff_info *info;
	int r;

	info = read_from_list(device_handle);
	if (!info) {
		_E("Handle %d fail to check info", device_handle);
		return -ENOENT; /* 2 */
	}

	if (!check_valid_handle(info)) {
		_E("Handle %d fail to check handle", device_handle);
		return -EINVAL; /* 22 */
	}

	if (!check_fd(&ff_fd)) {
		_E("Handle %d fail to check fd", device_handle);
		return -ENODEV; /* 19 */
	}

	if (cur_h_data.handle > 0 && cur_h_data.handle != info->handle) {
		_E("Only same handle can stop current vibration");
		return -EPERM;
	}

	/* stop effect */
	r = ff_stop(ff_fd, &info->effect);
	if (r < 0)
		_E("failed to stop effect(id:%d) : %d", info->effect.id, r);
	else
		_D("succeed to stop effect");

	/* unregister existing timer */
	if (r >= 0 && info->timer) {
		g_source_remove(info->timer);
		info->timer = 0;
	}

	return 0;
}
/* END: Haptic Module APIs */

static const struct haptic_plugin_ops default_plugin = {
	.get_device_count    = get_device_count,
	.open_device         = open_device,
	.close_device        = close_device,
	.vibrate_monotone    = vibrate_monotone,
	.stop_device         = stop_device,
};

static bool is_valid(void)
{
	int ret;

	ret = ff_find_device();
	if (ret < 0) {
		_E("Do not support standard haptic device");
		return false;
	}

	_I("Support standard haptic device");
	return true;
}

static const struct haptic_plugin_ops *load(void)
{
	return &default_plugin;
}

static const struct haptic_ops std_ops = {
	.type     = HAPTIC_STANDARD,
	.is_valid = is_valid,
	.load     = load,
};

HAPTIC_OPS_REGISTER(&std_ops)
